@@ -1,3 +1,17 @@
+# Copyright 2015 WebAssembly Community Group participants
+#
+# Licensed under the Apache License, Version 2.0 (the "License");
+# you may not use this file except in compliance with the License.
+# You may obtain a copy of the License at
+#
+#     http://www.apache.org/licenses/LICENSE-2.0
+#
+# Unless required by applicable law or agreed to in writing, software
+# distributed under the License is distributed on an "AS IS" BASIS,
+# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
+# See the License for the specific language governing permissions and
+# limitations under the License.
+
 import argparse
 import difflib
 import glob
@@ -5,79 +19,68 @@ import os
 import shutil
 import subprocess
 import sys
-import urllib2
 
 
-usage_str = ("usage: 'python check.py [options]'\n\n"
-             "Runs the Binaryen test suite.")
-parser = argparse.ArgumentParser(description=usage_str)
-parser.add_argument(
-    '--torture', dest='torture', action='store_true', default=True,
-    help='Chooses whether to run the torture testcases. Default: true.')
-parser.add_argument(
-    '--no-torture', dest='torture', action='store_false',
-    help='Disables running the torture testcases.')
-parser.add_argument(
-    '--only-prepare', dest='only_prepare', action='store_true', default=False,
-    help='If enabled, only fetches the waterfall build. Default: false.')
-parser.add_argument(
-    # Backwards compatibility
-    '--only_prepare', dest='only_prepare', action='store_true', default=False,
-    help='If enabled, only fetches the waterfall build. Default: false.')
-parser.add_argument(
-    '--test-waterfall', dest='test_waterfall', action='store_true',
-    default=False,
-    help=('If enabled, fetches and tests the LLVM waterfall builds.'
-          ' Default: false.'))
-parser.add_argument(
-    '--no-test-waterfall', dest='test_waterfall', action='store_false',
-    help='Disables downloading and testing of the LLVM waterfall builds.')
-parser.add_argument(
-    '--abort-on-first-failure', dest='abort_on_first_failure',
-    action='store_true', default=True,
-    help=('Specifies whether to halt test suite execution on first test error.'
-          ' Default: true.'))
-parser.add_argument(
-    '--no-abort-on-first-failure', dest='abort_on_first_failure',
-    action='store_false',
-    help=('If set, the whole test suite will run to completion independent of'
-          ' earlier errors.'))
-parser.add_argument(
-    '--run-gcc-tests', dest='run_gcc_tests', action='store_true', default=True,
-    help=('Chooses whether to run the tests that require building with native'
-          ' GCC. Default: true.'))
-parser.add_argument(
-    '--no-run-gcc-tests', dest='run_gcc_tests', action='store_false',
-    help='If set, disables the native GCC tests.')
+def parse_args(args):
+  usage_str = ("usage: 'python check.py [options]'\n\n"
+               "Runs the Binaryen test suite.")
+  parser = argparse.ArgumentParser(description=usage_str)
+  parser.add_argument(
+      '--torture', dest='torture', action='store_true', default=True,
+      help='Chooses whether to run the torture testcases. Default: true.')
+  parser.add_argument(
+      '--no-torture', dest='torture', action='store_false',
+      help='Disables running the torture testcases.')
+  parser.add_argument(
+      '--abort-on-first-failure', dest='abort_on_first_failure',
+      action='store_true', default=True,
+      help=('Specifies whether to halt test suite execution on first test error.'
+            ' Default: true.'))
+  parser.add_argument(
+      '--no-abort-on-first-failure', dest='abort_on_first_failure',
+      action='store_false',
+      help=('If set, the whole test suite will run to completion independent of'
+            ' earlier errors.'))
+  parser.add_argument(
+      '--run-gcc-tests', dest='run_gcc_tests', action='store_true', default=True,
+      help=('Chooses whether to run the tests that require building with native'
+            ' GCC. Default: true.'))
+  parser.add_argument(
+      '--no-run-gcc-tests', dest='run_gcc_tests', action='store_false',
+      help='If set, disables the native GCC tests.')
 
-parser.add_argument(
-    '--interpreter', dest='interpreter', default='',
-    help='Specifies the wasm interpreter executable to run tests on.')
-parser.add_argument(
-    '--binaryen-bin', dest='binaryen_bin', default='',
-    help=('Specifies a path to where the built Binaryen executables reside at.'
-          ' Default: bin/ of current directory (i.e. assume an in-tree build).'
-          ' If not specified, the environment variable BINARYEN_ROOT= can also'
-          ' be used to adjust this.'))
-parser.add_argument(
-    '--binaryen-root', dest='binaryen_root', default='',
-    help=('Specifies a path to the root of the Binaryen repository tree.'
-          ' Default: the directory where this file check.py resides.'))
-parser.add_argument(
-    '--valgrind', dest='valgrind', default='',
-    help=('Specifies a path to Valgrind tool, which will be used to validate'
-          ' execution if specified. (Pass --valgrind=valgrind to search in'
-          ' PATH)'))
-parser.add_argument(
-    '--valgrind-full-leak-check', dest='valgrind_full_leak_check',
-    action='store_true', default=False,
-    help=('If specified, all unfreed (but still referenced) pointers at the'
-          ' end of execution are considered memory leaks. Default: disabled.'))
+  parser.add_argument(
+      '--interpreter', dest='interpreter', default='',
+      help='Specifies the wasm interpreter executable to run tests on.')
+  parser.add_argument(
+      '--binaryen-bin', dest='binaryen_bin', default='',
+      help=('Specifies a path to where the built Binaryen executables reside at.'
+            ' Default: bin/ of current directory (i.e. assume an in-tree build).'
+            ' If not specified, the environment variable BINARYEN_ROOT= can also'
+            ' be used to adjust this.'))
+  parser.add_argument(
+      '--binaryen-root', dest='binaryen_root', default='',
+      help=('Specifies a path to the root of the Binaryen repository tree.'
+            ' Default: the directory where this file check.py resides.'))
+  parser.add_argument(
+      '--valgrind', dest='valgrind', default='',
+      help=('Specifies a path to Valgrind tool, which will be used to validate'
+            ' execution if specified. (Pass --valgrind=valgrind to search in'
+            ' PATH)'))
+  parser.add_argument(
+      '--valgrind-full-leak-check', dest='valgrind_full_leak_check',
+      action='store_true', default=False,
+      help=('If specified, all unfreed (but still referenced) pointers at the'
+            ' end of execution are considered memory leaks. Default: disabled.'))
 
-parser.add_argument(
-    'positional_args', metavar='tests', nargs=argparse.REMAINDER,
-    help='Names specific tests to run.')
-options = parser.parse_args()
+  parser.add_argument(
+      'positional_args', metavar='tests', nargs=argparse.REMAINDER,
+      help='Names specific tests to run.')
+
+  return parser.parse_args(args)
+
+
+options = parse_args(sys.argv[1:])
 requested = options.positional_args
 
 num_failures = 0
@@ -163,18 +166,15 @@ WASM_OPT = [os.path.join(options.binaryen_bin, 'wasm-opt')]
 WASM_AS = [os.path.join(options.binaryen_bin, 'wasm-as')]
 WASM_DIS = [os.path.join(options.binaryen_bin, 'wasm-dis')]
 ASM2WASM = [os.path.join(options.binaryen_bin, 'asm2wasm')]
-WASM2ASM = [os.path.join(options.binaryen_bin, 'wasm2asm')]
+WASM2JS = [os.path.join(options.binaryen_bin, 'wasm2js')]
 WASM_CTOR_EVAL = [os.path.join(options.binaryen_bin, 'wasm-ctor-eval')]
 WASM_SHELL = [os.path.join(options.binaryen_bin, 'wasm-shell')]
 WASM_MERGE = [os.path.join(options.binaryen_bin, 'wasm-merge')]
-S2WASM = [os.path.join(options.binaryen_bin, 's2wasm')]
 WASM_REDUCE = [os.path.join(options.binaryen_bin, 'wasm-reduce')]
 WASM_METADCE = [os.path.join(options.binaryen_bin, 'wasm-metadce')]
 WASM_EMSCRIPTEN_FINALIZE = [os.path.join(options.binaryen_bin,
                                          'wasm-emscripten-finalize')]
-
-S2WASM_EXE = S2WASM[0]
-WASM_SHELL_EXE = WASM_SHELL[0]
+BINARYEN_JS = os.path.join(options.binaryen_bin, 'binaryen.js')
 
 
 def wrap_with_valgrind(cmd):
@@ -192,7 +192,6 @@ if options.valgrind:
   WASM_DIS = wrap_with_valgrind(WASM_DIS)
   ASM2WASM = wrap_with_valgrind(ASM2WASM)
   WASM_SHELL = wrap_with_valgrind(WASM_SHELL)
-  S2WASM = wrap_with_valgrind(S2WASM)
 
 os.environ['BINARYEN'] = os.getcwd()
 
@@ -208,59 +207,8 @@ def has_shell_timeout():
   return get_platform() != 'windows' and os.system('timeout 1s pwd') == 0
 
 
-def fetch_waterfall():
-  rev = open(os.path.join(options.binaryen_test, 'revision')).read().strip()
-  buildername = get_platform()
-  local_rev_path = os.path.join(WATERFALL_BUILD_DIR, 'local-revision')
-  if os.path.exists(local_rev_path):
-    with open(local_rev_path) as f:
-      local_rev = f.read().strip()
-    if local_rev == rev:
-      return
-  # fetch it
-  basename = 'wasm-binaries-' + rev + '.tbz2'
-  url = '/'.join(['https://storage.googleapis.com/wasm-llvm/builds',
-                  buildername, rev, basename])
-  print '(downloading waterfall %s: %s)' % (rev, url)
-  downloaded = urllib2.urlopen(url).read().strip()
-  fullname = os.path.join(options.binaryen_test, basename)
-  open(fullname, 'wb').write(downloaded)
-  print '(unpacking)'
-  if os.path.exists(WATERFALL_BUILD_DIR):
-    shutil.rmtree(WATERFALL_BUILD_DIR)
-  os.mkdir(WATERFALL_BUILD_DIR)
-  subprocess.check_call(['tar', '-xf', os.path.abspath(fullname)],
-                        cwd=WATERFALL_BUILD_DIR)
-  print '(noting local revision)'
-  with open(local_rev_path, 'w') as o:
-    o.write(rev + '\n')
-
-
 has_vanilla_llvm = False
 
-
-def setup_waterfall():
-  # if we can use the waterfall llvm, do so
-  global has_vanilla_llvm
-  CLANG = os.path.join(BIN_DIR, 'clang')
-  print 'trying waterfall clang at', CLANG
-  try:
-    subprocess.check_call([CLANG, '-v'])
-    has_vanilla_llvm = True
-    print '...success'
-  except (OSError, subprocess.CalledProcessError) as e:
-    warn('could not run vanilla LLVM from waterfall: ' + str(e) +
-         ', looked for clang at ' + CLANG)
-
-
-if options.test_waterfall:
-  fetch_waterfall()
-  setup_waterfall()
-
-if options.only_prepare:
-  print 'waterfall is fetched and setup, exiting since --only-prepare'
-  sys.exit(0)
-
 # external tools
 
 try:
@@ -363,7 +311,7 @@ def fail_if_not_contained(actual, expected):
 
 
 def fail_if_not_identical_to_file(actual, expected_file):
-  with open(expected_file, 'rb') as f:
+  with open(expected_file, 'rb' if expected_file.endswith(".wasm") else 'r') as f:
     fail_if_not_identical(actual, f.read(), fromfile=expected_file)
 
 
