@@ -27,8 +27,9 @@
 #include "emscripten-optimizer/istring.h"
 #include "support/name.h"
 #include "wasm-builder.h"
+#include "ir/module-utils.h"
 #include "ir/names.h"
-
+#include "asmjs/shared-constants.h"
 
 namespace wasm {
 
@@ -40,10 +41,10 @@ static Name makeHighName(Name n) {
 
 struct I64ToI32Lowering : public WalkerPass<PostWalker<I64ToI32Lowering>> {
   struct TempVar {
-    TempVar(Index idx, I64ToI32Lowering& pass) :
-      idx(idx), pass(pass), moved(false) {}
+    TempVar(Index idx, Type ty, I64ToI32Lowering& pass) :
+      idx(idx), pass(pass), moved(false), ty(ty) {}
 
-    TempVar(TempVar&& other) : idx(other), pass(other.pass), moved(false) {
+    TempVar(TempVar&& other) : idx(other), pass(other.pass), moved(false), ty(other.ty) {
       assert(!other.moved);
       other.moved = true;
     }
@@ -78,18 +79,17 @@ struct I64ToI32Lowering : public WalkerPass<PostWalker<I64ToI32Lowering>> {
 
   private:
     void freeIdx() {
-      assert(std::find(pass.freeTemps.begin(), pass.freeTemps.end(), idx) ==
-             pass.freeTemps.end());
-      pass.freeTemps.push_back(idx);
+      auto &freeList = pass.freeTemps[(int) ty];
+      assert(std::find(freeList.begin(), freeList.end(), idx) == freeList.end());
+      freeList.push_back(idx);
     }
 
     Index idx;
     I64ToI32Lowering& pass;
     bool moved; // since C++ will still destruct moved-from values
+    Type ty;
   };
 
-  static Name highBitsGlobal;
-
   // false since function types need to be lowered
   // TODO: allow module-level transformations in parallel passes
   bool isFunctionParallel() override { return false; }
@@ -105,7 +105,7 @@ struct I64ToI32Lowering : public WalkerPass<PostWalker<I64ToI32Lowering>> {
       auto& curr = module->globals[i];
       if (curr->type != i64) continue;
       curr->type = i32;
-      auto* high = new Global(*curr);
+      auto* high = ModuleUtils::copyGlobal(curr.get(), *module);
       high->name = makeHighName(curr->name);
       module->addGlobal(high);
     }
@@ -114,7 +114,7 @@ struct I64ToI32Lowering : public WalkerPass<PostWalker<I64ToI32Lowering>> {
     // to return the high 32 bits.
     auto* highBits = new Global();
     highBits->type = i32;
-    highBits->name = highBitsGlobal;
+    highBits->name = INT64_TO_32_HIGH_BITS;
     highBits->init = builder->makeConst(Literal(int32_t(0)));
     highBits->mutable_ = true;
     module->addGlobal(highBits);
@@ -144,19 +144,20 @@ struct I64ToI32Lowering : public WalkerPass<PostWalker<I64ToI32Lowering>> {
     highBitVars.clear();
     labelHighBitVars.clear();
     freeTemps.clear();
-    Function oldFunc(*func);
+    Module temp;
+    auto* oldFunc = ModuleUtils::copyFunction(func, temp);
     func->params.clear();
     func->vars.clear();
     func->localNames.clear();
     func->localIndices.clear();
     Index newIdx = 0;
-    Names::ensureNames(&oldFunc);
-    for (Index i = 0; i < oldFunc.getNumLocals(); ++i) {
-      assert(oldFunc.hasLocalName(i));
-      Name lowName = oldFunc.getLocalName(i);
+    Names::ensureNames(oldFunc);
+    for (Index i = 0; i < oldFunc->getNumLocals(); ++i) {
+      assert(oldFunc->hasLocalName(i));
+      Name lowName = oldFunc->getLocalName(i);
       Name highName = makeHighName(lowName);
-      Type paramType = oldFunc.getLocalType(i);
-      auto builderFunc = (i < oldFunc.getVarIndexBase()) ?
+      Type paramType = oldFunc->getLocalType(i);
+      auto builderFunc = (i < oldFunc->getVarIndexBase()) ?
           Builder::addParam :
           static_cast<Index (*)(Function*, Name, Type)>(Builder::addVar);
       if (paramType == i64) {
@@ -174,6 +175,9 @@ struct I64ToI32Lowering : public WalkerPass<PostWalker<I64ToI32Lowering>> {
   }
 
   void visitFunction(Function* func) {
+    if (func->imported()) {
+      return;
+    }
     if (func->result == i64) {
       func->result = i32;
       // body may not have out param if it ends with control flow
@@ -185,18 +189,17 @@ struct I64ToI32Lowering : public WalkerPass<PostWalker<I64ToI32Lowering>> {
           func->body
         );
         SetGlobal* setHigh = builder->makeSetGlobal(
-          highBitsGlobal,
+          INT64_TO_32_HIGH_BITS,
           builder->makeGetLocal(highBits, i32)
         );
         GetLocal* getLow = builder->makeGetLocal(lowBits, i32);
         func->body = builder->blockify(setLow, setHigh, getLow);
       }
     }
-    assert(freeTemps.size() == nextTemp - func->getNumLocals());
     int idx = 0;
     for (size_t i = func->getNumLocals(); i < nextTemp; i++) {
       Name tmpName("i64toi32_i32$" + std::to_string(idx++));
-      builder->addVar(func, tmpName, i32);
+      builder->addVar(func, tmpName, tempTypes[i]);
     }
   }
 
@@ -345,7 +348,7 @@ struct I64ToI32Lowering : public WalkerPass<PostWalker<I64ToI32Lowering>> {
     );
     SetLocal* setHigh = builder->makeSetLocal(
       highBits,
-      builder->makeGetGlobal(highBitsGlobal, i32)
+      builder->makeGetGlobal(INT64_TO_32_HIGH_BITS, i32)
     );
     GetLocal* getLow = builder->makeGetLocal(lowBits, i32);
     Block* result = builder->blockify(doCall, setHigh, getLow);
@@ -361,11 +364,6 @@ struct I64ToI32Lowering : public WalkerPass<PostWalker<I64ToI32Lowering>> {
     );
   }
 
-  void visitCallImport(CallImport* curr) {
-    // imports cannot contain i64s
-    return;
-  }
-
   void visitCallIndirect(CallIndirect* curr) {
     visitGenericCall<CallIndirect>(
       curr,
@@ -451,6 +449,7 @@ struct I64ToI32Lowering : public WalkerPass<PostWalker<I64ToI32Lowering>> {
   void visitLoad(Load* curr) {
     if (curr->type != i64) return;
     assert(!curr->isAtomic && "atomic load not implemented");
+    TempVar lowBits = getTemp();
     TempVar highBits = getTemp();
     TempVar ptrTemp = getTemp();
     SetLocal* setPtr = builder->makeSetLocal(ptrTemp, curr->ptr);
@@ -467,6 +466,15 @@ struct I64ToI32Lowering : public WalkerPass<PostWalker<I64ToI32Lowering>> {
           i32
         )
       );
+    } else if (curr->signed_) {
+      loadHigh = builder->makeSetLocal(
+        highBits,
+        builder->makeBinary(
+          ShrSInt32,
+          builder->makeGetLocal(lowBits, i32),
+          builder->makeConst(Literal(int32_t(31)))
+        )
+      );
     } else {
       loadHigh = builder->makeSetLocal(
         highBits,
@@ -477,7 +485,12 @@ struct I64ToI32Lowering : public WalkerPass<PostWalker<I64ToI32Lowering>> {
     curr->bytes = std::min(curr->bytes, uint8_t(4));
     curr->align = std::min(uint32_t(curr->align), uint32_t(4));
     curr->ptr = builder->makeGetLocal(ptrTemp, i32);
-    Block* result = builder->blockify(setPtr, loadHigh, curr);
+    Block* result = builder->blockify(
+      setPtr,
+      builder->makeSetLocal(lowBits, curr),
+      loadHigh,
+      builder->makeGetLocal(lowBits, i32)
+    );
     replaceCurrent(result);
     setOutParam(result, std::move(highBits));
   }
@@ -587,28 +600,212 @@ struct I64ToI32Lowering : public WalkerPass<PostWalker<I64ToI32Lowering>> {
     replaceCurrent(curr->value);
   }
 
-  void lowerPopcnt64(Unary* curr) {
+  void lowerReinterpretFloat64(Unary* curr) {
+    // Assume that the wasm file assumes the address 0 is invalid and roundtrip
+    // our f64 through memory at address 0
+    TempVar highBits = getTemp();
+    Block *result = builder->blockify(
+      builder->makeStore(8, 0, 8, builder->makeConst(Literal(int32_t(0))), curr->value, f64),
+      builder->makeSetLocal(
+        highBits,
+        builder->makeLoad(4, true, 4, 4, builder->makeConst(Literal(int32_t(0))), i32)
+      ),
+      builder->makeLoad(4, true, 0, 4, builder->makeConst(Literal(int32_t(0))), i32)
+    );
+    setOutParam(result, std::move(highBits));
+    replaceCurrent(result);
+  }
+
+  void lowerReinterpretInt64(Unary* curr) {
+    // Assume that the wasm file assumes the address 0 is invalid and roundtrip
+    // our i64 through memory at address 0
+    TempVar highBits = fetchOutParam(curr->value);
+    Block *result = builder->blockify(
+      builder->makeStore(4, 0, 4, builder->makeConst(Literal(int32_t(0))), curr->value, i32),
+      builder->makeStore(4, 4, 4, builder->makeConst(Literal(int32_t(0))), builder->makeGetLocal(highBits, i32), i32),
+      builder->makeLoad(8, true, 0, 8, builder->makeConst(Literal(int32_t(0))), f64)
+    );
+    replaceCurrent(result);
+  }
+
+  void lowerTruncFloatToInt(Unary *curr) {
+    // hiBits = if abs(f) >= 1.0 {
+    //    if f > 0.0 {
+    //        (unsigned) min(
+    //          floor(f / (float) U32_MAX),
+    //          (float) U32_MAX - 1,
+    //        )
+    //    } else {
+    //        (unsigned) ceil((f - (float) (unsigned) f) / ((float) U32_MAX))
+    //    }
+    // } else {
+    //    0
+    // }
+    //
+    // loBits = (unsigned) f;
+
+    Literal litZero, litOne, u32Max;
+    UnaryOp trunc, convert, abs, floor, ceil;
+    Type localType;
+    BinaryOp ge, gt, min, div, sub;
+    switch (curr->op) {
+      case TruncSFloat32ToInt64:
+      case TruncUFloat32ToInt64: {
+        litZero = Literal((float) 0);
+        litOne = Literal((float) 1);
+        u32Max = Literal(((float) UINT_MAX) + 1);
+        trunc = TruncUFloat32ToInt32;
+        convert = ConvertUInt32ToFloat32;
+        localType = f32;
+        abs = AbsFloat32;
+        ge = GeFloat32;
+        gt = GtFloat32;
+        min = MinFloat32;
+        floor = FloorFloat32;
+        ceil = CeilFloat32;
+        div = DivFloat32;
+        sub = SubFloat32;
+        break;
+      }
+      case TruncSFloat64ToInt64:
+      case TruncUFloat64ToInt64: {
+        litZero = Literal((double) 0);
+        litOne = Literal((double) 1);
+        u32Max = Literal(((double) UINT_MAX) + 1);
+        trunc = TruncUFloat64ToInt32;
+        convert = ConvertUInt32ToFloat64;
+        localType = f64;
+        abs = AbsFloat64;
+        ge = GeFloat64;
+        gt = GtFloat64;
+        min = MinFloat64;
+        floor = FloorFloat64;
+        ceil = CeilFloat64;
+        div = DivFloat64;
+        sub = SubFloat64;
+        break;
+      }
+      default: abort();
+    }
+
+    TempVar f = getTemp(localType);
+    TempVar highBits = getTemp();
+
+    Expression *gtZeroBranch = builder->makeBinary(
+        min,
+        builder->makeUnary(
+          floor,
+          builder->makeBinary(
+            div,
+            builder->makeGetLocal(f, localType),
+            builder->makeConst(u32Max)
+          )
+        ),
+        builder->makeBinary(sub, builder->makeConst(u32Max), builder->makeConst(litOne))
+    );
+    Expression *ltZeroBranch = builder->makeUnary(
+        ceil,
+        builder->makeBinary(
+          div,
+          builder->makeBinary(
+            sub,
+            builder->makeGetLocal(f, localType),
+            builder->makeUnary(convert,
+              builder->makeUnary(trunc, builder->makeGetLocal(f, localType))
+            )
+          ),
+          builder->makeConst(u32Max)
+        )
+    );
+
+    If *highBitsCalc = builder->makeIf(
+      builder->makeBinary(
+        gt,
+        builder-> makeGetLocal(f, localType),
+        builder->makeConst(litZero)
+      ),
+      builder->makeUnary(trunc, gtZeroBranch),
+      builder->makeUnary(trunc, ltZeroBranch)
+    );
+    If *highBitsVal = builder->makeIf(
+      builder->makeBinary(
+        ge,
+        builder->makeUnary(abs, builder->makeGetLocal(f, localType)),
+        builder->makeConst(litOne)
+      ),
+      highBitsCalc,
+      builder->makeConst(Literal(int32_t(0)))
+    );
+    Block *result = builder->blockify(
+      builder->makeSetLocal(f, curr->value),
+      builder->makeSetLocal(highBits, highBitsVal),
+      builder->makeUnary(trunc, builder->makeGetLocal(f, localType))
+    );
+    setOutParam(result, std::move(highBits));
+    replaceCurrent(result);
+  }
+
+  void lowerConvertIntToFloat(Unary *curr) {
+    // Here the same strategy as `emcc` is taken which takes the two halves of
+    // the 64-bit integer and creates a mathematical expression using float
+    // arithmetic to reassemble the final floating point value.
+    //
+    // For example for i64 -> f32 we generate:
+    //
+    //  ((double) (unsigned) lowBits) + ((double) U32_MAX) * ((double) (int) highBits)
+    //
+    // Mostly just shuffling things around here with coercions and whatnot!
+    // Note though that all arithmetic is done with f64 to have as much
+    // precision as we can.
     TempVar highBits = fetchOutParam(curr->value);
     TempVar lowBits = getTemp();
     TempVar highResult = getTemp();
 
-    SetLocal* setLow = builder->makeSetLocal(lowBits, curr->value);
-    SetLocal* setHigh = builder->makeSetLocal(
-      highResult,
-      builder->makeConst(Literal(int32_t(0)))
-    );
+    UnaryOp convertHigh;
+    switch (curr->op) {
+      case ConvertSInt64ToFloat32:
+      case ConvertSInt64ToFloat64:
+        convertHigh = ConvertSInt32ToFloat64;
+        break;
+      case ConvertUInt64ToFloat32:
+      case ConvertUInt64ToFloat64:
+        convertHigh = ConvertUInt32ToFloat64;
+        break;
+      default: abort();
+    }
 
-    Block* result = builder->blockify(
-      setLow,
-      setHigh,
+    Expression *result = builder->blockify(
+      builder->makeSetLocal(lowBits, curr->value),
+      builder->makeSetLocal(
+        highResult,
+        builder->makeConst(Literal(int32_t(0)))
+      ),
       builder->makeBinary(
-        AddInt32,
-        builder->makeUnary(PopcntInt32, builder->makeGetLocal(highBits, i32)),
-        builder->makeUnary(PopcntInt32, builder->makeGetLocal(lowBits, i32))
+        AddFloat64,
+        builder->makeUnary(
+          ConvertUInt32ToFloat64,
+          builder->makeGetLocal(lowBits, i32)
+        ),
+        builder->makeBinary(
+          MulFloat64,
+          builder->makeConst(Literal((double)UINT_MAX + 1)),
+          builder->makeUnary(
+            convertHigh,
+            builder->makeGetLocal(highBits, i32)
+          )
+        )
       )
     );
 
-    setOutParam(result, std::move(highResult));
+    switch (curr->op) {
+      case ConvertSInt64ToFloat32:
+      case ConvertUInt64ToFloat32: {
+        result = builder->makeUnary(DemoteFloat64, result);
+        break;
+      }
+      default: break;
+    }
+
     replaceCurrent(result);
   }
 
@@ -664,7 +861,8 @@ struct I64ToI32Lowering : public WalkerPass<PostWalker<I64ToI32Lowering>> {
         lower(result, ClzInt32, std::move(highBits), std::move(lowBits));
         break;
       case CtzInt64:
-        lower(result, CtzInt32, std::move(lowBits), std::move(highBits));
+        std::cerr << "i64.ctz should be removed already" << std::endl;
+        WASM_UNREACHABLE();
         break;
       default:
         abort();
@@ -701,25 +899,27 @@ struct I64ToI32Lowering : public WalkerPass<PostWalker<I64ToI32Lowering>> {
       replaceCurrent(curr->value);
       return;
     }
-    assert(hasOutParam(curr->value) || curr->type == i64);
+    assert(hasOutParam(curr->value) || curr->type == i64 || curr->type == f64);
     switch (curr->op) {
       case ClzInt64:
       case CtzInt64:               lowerCountZeros(curr);   break;
-      case PopcntInt64:            lowerPopcnt64(curr);     break;
       case EqZInt64:               lowerEqZInt64(curr);     break;
       case ExtendSInt32:           lowerExtendSInt32(curr); break;
       case ExtendUInt32:           lowerExtendUInt32(curr); break;
       case WrapInt64:              lowerWrapInt64(curr);    break;
+      case ReinterpretFloat64:     lowerReinterpretFloat64(curr); break;
+      case ReinterpretInt64:       lowerReinterpretInt64(curr);   break;
       case TruncSFloat32ToInt64:
       case TruncUFloat32ToInt64:
       case TruncSFloat64ToInt64:
-      case TruncUFloat64ToInt64:
-      case ReinterpretFloat64:
+      case TruncUFloat64ToInt64:   lowerTruncFloatToInt(curr);   break;
       case ConvertSInt64ToFloat32:
       case ConvertSInt64ToFloat64:
       case ConvertUInt64ToFloat32:
-      case ConvertUInt64ToFloat64:
-      case ReinterpretInt64:
+      case ConvertUInt64ToFloat64: lowerConvertIntToFloat(curr); break;
+      case PopcntInt64:
+        std::cerr << "i64.popcnt should already be removed" << std::endl;
+        WASM_UNREACHABLE();
       default:
         std::cerr << "Unhandled unary operator: " << curr->op << std::endl;
         abort();
@@ -811,122 +1011,6 @@ struct I64ToI32Lowering : public WalkerPass<PostWalker<I64ToI32Lowering>> {
     return result;
   }
 
-  Block* lowerMul(Block* result, TempVar&& leftLow, TempVar&& leftHigh,
-                  TempVar&& rightLow, TempVar&& rightHigh) {
-    // high bits = ll*rh + lh*rl + ll1*rl1 + (ll0*rl1)>>16 + (ll1*rl0)>>16
-    // low bits = ll*rl
-    TempVar leftLow0 = getTemp();
-    TempVar leftLow1 = getTemp();
-    TempVar rightLow0 = getTemp();
-    TempVar rightLow1 = getTemp();
-    SetLocal* setLL0 = builder->makeSetLocal(
-      leftLow0,
-      builder->makeBinary(
-        AndInt32,
-        builder->makeGetLocal(leftLow, i32),
-        builder->makeConst(Literal(int32_t(0xffff)))
-      )
-    );
-    SetLocal* setLL1 = builder->makeSetLocal(
-      leftLow1,
-      builder->makeBinary(
-        ShrUInt32,
-        builder->makeGetLocal(leftLow, i32),
-        builder->makeConst(Literal(int32_t(16)))
-      )
-    );
-    SetLocal* setRL0 = builder->makeSetLocal(
-      rightLow0,
-      builder->makeBinary(
-        AndInt32,
-        builder->makeGetLocal(rightLow, i32),
-        builder->makeConst(Literal(int32_t(0xffff)))
-      )
-    );
-    SetLocal* setRL1 = builder->makeSetLocal(
-      rightLow1,
-      builder->makeBinary(
-        ShrUInt32,
-        builder->makeGetLocal(rightLow, i32),
-        builder->makeConst(Literal(int32_t(16)))
-      )
-    );
-    SetLocal* setLLRH = builder->makeSetLocal(
-      rightHigh,
-      builder->makeBinary(
-        MulInt32,
-        builder->makeGetLocal(leftLow, i32),
-        builder->makeGetLocal(rightHigh, i32)
-      )
-    );
-    auto addToHighBits = [&](Expression* expr) -> SetLocal* {
-      return builder->makeSetLocal(
-        rightHigh,
-        builder->makeBinary(
-          AddInt32,
-          builder->makeGetLocal(rightHigh, i32),
-          expr
-        )
-      );
-    };
-    SetLocal* addLHRL = addToHighBits(
-      builder->makeBinary(
-        MulInt32,
-        builder->makeGetLocal(leftHigh, i32),
-        builder->makeGetLocal(rightLow, i32)
-      )
-    );
-    SetLocal* addLL1RL1 = addToHighBits(
-      builder->makeBinary(
-        MulInt32,
-        builder->makeGetLocal(leftLow1, i32),
-        builder->makeGetLocal(rightLow1, i32)
-      )
-    );
-    SetLocal* addLL0RL1 = addToHighBits(
-      builder->makeBinary(
-        ShrUInt32,
-        builder->makeBinary(
-          MulInt32,
-          builder->makeGetLocal(leftLow0, i32),
-          builder->makeGetLocal(rightLow1, i32)
-        ),
-        builder->makeConst(Literal(int32_t(16)))
-      )
-    );
-    SetLocal* addLL1RL0 = addToHighBits(
-      builder->makeBinary(
-        ShrUInt32,
-        builder->makeBinary(
-          MulInt32,
-          builder->makeGetLocal(leftLow1, i32),
-          builder->makeGetLocal(rightLow0, i32)
-        ),
-        builder->makeConst(Literal(int32_t(16)))
-      )
-    );
-    Binary* getLow = builder->makeBinary(
-      MulInt32,
-      builder->makeGetLocal(leftLow, i32),
-      builder->makeGetLocal(rightLow, i32)
-    );
-    result = builder->blockify(
-      result,
-      setLL0,
-      setLL1,
-      setRL0,
-      setRL1,
-      setLLRH,
-      addLHRL,
-      addLL1RL1,
-      addLL0RL1,
-      addLL1RL0,
-      getLow
-    );
-    setOutParam(result, std::move(rightHigh));
-    return result;
-  }
-
   Block* lowerBitwise(BinaryOp op, Block* result, TempVar&& leftLow,
                       TempVar&& leftHigh, TempVar&& rightLow,
                       TempVar&& rightHigh) {
@@ -971,6 +1055,30 @@ struct I64ToI32Lowering : public WalkerPass<PostWalker<I64ToI32Lowering>> {
     );
   }
 
+  // a >> b where `b` >= 32
+  //
+  // implement as:
+  //
+  // hi = leftHigh >> 31 // copy sign bit
+  // lo = leftHigh >> (b - 32)
+  Block* makeLargeShrS(Index highBits, Index leftHigh, Index shift) {
+    return builder->blockify(
+      builder->makeSetLocal(
+        highBits,
+        builder->makeBinary(
+          ShrSInt32,
+          builder->makeGetLocal(leftHigh, i32),
+          builder->makeConst(Literal(int32_t(31)))
+        )
+      ),
+      builder->makeBinary(
+        ShrSInt32,
+        builder->makeGetLocal(leftHigh, i32),
+        builder->makeGetLocal(shift, i32)
+      )
+    );
+  }
+
   Block* makeLargeShrU(Index highBits, Index leftHigh, Index shift) {
     return builder->blockify(
       builder->makeSetLocal(highBits, builder->makeConst(Literal(int32_t(0)))),
@@ -1011,6 +1119,41 @@ struct I64ToI32Lowering : public WalkerPass<PostWalker<I64ToI32Lowering>> {
     );
   }
 
+  // a >> b where `b` < 32
+  //
+  // implement as:
+  //
+  // hi = leftHigh >> b
+  // lo = (leftLow >>> b) | (leftHigh << (32 - b))
+  Block* makeSmallShrS(Index highBits, Index leftLow, Index leftHigh,
+                       Index shift, Binary* shiftMask, Binary* widthLessShift) {
+    Binary* shiftedInBits = builder->makeBinary(
+      ShlInt32,
+      builder->makeBinary(
+        AndInt32,
+        shiftMask,
+        builder->makeGetLocal(leftHigh, i32)
+      ),
+      widthLessShift
+    );
+    Binary* shiftLow = builder->makeBinary(
+      ShrUInt32,
+      builder->makeGetLocal(leftLow, i32),
+      builder->makeGetLocal(shift, i32)
+    );
+    return builder->blockify(
+      builder->makeSetLocal(
+        highBits,
+        builder->makeBinary(
+          ShrSInt32,
+          builder->makeGetLocal(leftHigh, i32),
+          builder->makeGetLocal(shift, i32)
+        )
+      ),
+      builder->makeBinary(OrInt32, shiftedInBits, shiftLow)
+    );
+  }
+
   Block* makeSmallShrU(Index highBits, Index leftLow, Index leftHigh,
                        Index shift, Binary* shiftMask, Binary* widthLessShift) {
     Binary* shiftedInBits = builder->makeBinary(
@@ -1040,9 +1183,9 @@ struct I64ToI32Lowering : public WalkerPass<PostWalker<I64ToI32Lowering>> {
     );
   }
 
-  Block* lowerShU(BinaryOp op, Block* result, TempVar&& leftLow,
-                  TempVar&& leftHigh, TempVar&& rightLow, TempVar&& rightHigh) {
-    assert(op == ShlInt64 || op == ShrUInt64);
+  Block* lowerShift(BinaryOp op, Block* result, TempVar&& leftLow,
+                    TempVar&& leftHigh, TempVar&& rightLow, TempVar&& rightHigh) {
+    assert(op == ShlInt64 || op == ShrUInt64 || op == ShrSInt64);
     // shift left lowered as:
     // if 32 <= rightLow % 64:
     //     high = leftLow << k; low = 0
@@ -1072,6 +1215,8 @@ struct I64ToI32Lowering : public WalkerPass<PostWalker<I64ToI32Lowering>> {
     switch (op) {
       case ShlInt64:
         largeShiftBlock = makeLargeShl(rightHigh, leftLow, shift); break;
+      case ShrSInt64:
+        largeShiftBlock = makeLargeShrS(rightHigh, leftHigh, shift); break;
       case ShrUInt64:
         largeShiftBlock = makeLargeShrU(rightHigh, leftHigh, shift); break;
       default: abort();
@@ -1097,6 +1242,11 @@ struct I64ToI32Lowering : public WalkerPass<PostWalker<I64ToI32Lowering>> {
                                        shift, shiftMask, widthLessShift);
         break;
       }
+      case ShrSInt64: {
+        smallShiftBlock = makeSmallShrS(rightHigh, leftLow, leftHigh,
+                                        shift, shiftMask, widthLessShift);
+        break;
+      }
       case ShrUInt64: {
         smallShiftBlock = makeSmallShrU(rightHigh, leftLow, leftHigh,
                                         shift, shiftMask, widthLessShift);
@@ -1306,17 +1456,16 @@ struct I64ToI32Lowering : public WalkerPass<PostWalker<I64ToI32Lowering>> {
                    std::move(rightLow), std::move(rightHigh)));
         break;
       }
-      case MulInt64: {
-        replaceCurrent(
-          lowerMul(result, std::move(leftLow), std::move(leftHigh),
-                   std::move(rightLow), std::move(rightHigh))
-        );
-        break;
-      }
+      case MulInt64:
       case DivSInt64:
       case DivUInt64:
       case RemSInt64:
-      case RemUInt64: goto err;
+      case RemUInt64:
+      case RotLInt64:
+      case RotRInt64:
+        std::cerr << "should have been removed by now " << curr->op << std::endl;
+        WASM_UNREACHABLE();
+
       case AndInt64:
       case OrInt64:
       case XorInt64: {
@@ -1328,16 +1477,14 @@ struct I64ToI32Lowering : public WalkerPass<PostWalker<I64ToI32Lowering>> {
         break;
       }
       case ShlInt64:
+      case ShrSInt64:
       case ShrUInt64: {
         replaceCurrent(
-          lowerShU(curr->op, result, std::move(leftLow), std::move(leftHigh),
-                   std::move(rightLow), std::move(rightHigh))
+          lowerShift(curr->op, result, std::move(leftLow), std::move(leftHigh),
+                     std::move(rightLow), std::move(rightHigh))
         );
         break;
       }
-      case ShrSInt64:
-      case RotLInt64:
-      case RotRInt64: goto err;
       case EqInt64: {
         replaceCurrent(
           lowerEq(result, std::move(leftLow), std::move(leftHigh),
@@ -1371,7 +1518,7 @@ struct I64ToI32Lowering : public WalkerPass<PostWalker<I64ToI32Lowering>> {
         );
         break;
       }
-      err: default: {
+      default: {
         std::cerr << "Unhandled binary op " << curr->op << std::endl;
         abort();
       }
@@ -1394,7 +1541,7 @@ struct I64ToI32Lowering : public WalkerPass<PostWalker<I64ToI32Lowering>> {
     TempVar highBits = fetchOutParam(curr->value);
     SetLocal* setLow = builder->makeSetLocal(lowBits, curr->value);
     SetGlobal* setHigh = builder->makeSetGlobal(
-      highBitsGlobal,
+      INT64_TO_32_HIGH_BITS,
       builder->makeGetLocal(highBits, i32)
     );
     curr->value = builder->makeGetLocal(lowBits, i32);
@@ -1405,20 +1552,24 @@ struct I64ToI32Lowering : public WalkerPass<PostWalker<I64ToI32Lowering>> {
 private:
   std::unique_ptr<Builder> builder;
   std::unordered_map<Index, Index> indexMap;
+  std::unordered_map<int, std::vector<Index>> freeTemps;
   std::unordered_map<Expression*, TempVar> highBitVars;
   std::unordered_map<Name, TempVar> labelHighBitVars;
-  std::vector<Index> freeTemps;
+  std::unordered_map<Index, Type> tempTypes;
   Index nextTemp;
 
-  TempVar getTemp() {
+  TempVar getTemp(Type ty = i32) {
     Index ret;
-    if (freeTemps.size() > 0) {
-      ret = freeTemps.back();
-      freeTemps.pop_back();
+    auto &freeList = freeTemps[(int) ty];
+    if (freeList.size() > 0) {
+      ret = freeList.back();
+      freeList.pop_back();
     } else {
       ret = nextTemp++;
+      tempTypes[ret] = ty;
     }
-    return TempVar(ret, *this);
+    assert(tempTypes[ret] == ty);
+    return TempVar(ret, ty, *this);
   }
 
   bool hasOutParam(Expression* e) {
@@ -1438,8 +1589,6 @@ private:
   }
 };
 
-Name I64ToI32Lowering::highBitsGlobal("i64toi32_i32$HIGH_BITS");
-
 Pass *createI64ToI32LoweringPass() {
   return new I64ToI32Lowering();
 }
