@@ -23,6 +23,7 @@
 #include "asm_v_wasm.h"
 #include "asmjs/shared-constants.h"
 #include "ir/branch-utils.h"
+#include "ir/function-type-utils.h"
 #include "shared-constants.h"
 #include "wasm-binary.h"
 #include "wasm-builder.h"
@@ -80,10 +81,10 @@ Element* Element::setString(IString str__, bool dollared__, bool quoted__) {
   return this;
 }
 
-Element* Element::setMetadata(size_t line_, size_t col_, SourceLocation* loc_) {
+Element* Element::setMetadata(size_t line_, size_t col_, SourceLocation* startLoc_) {
   line = line_;
   col = col_;
-  loc = loc_;
+  startLoc = startLoc_;
   return this;
 }
 
@@ -127,6 +128,7 @@ Element* SExpressionParser::parse() {
       assert(stack.size() == stackLocs.size());
     } else if (input[0] == ')') {
       input++;
+      curr->endLoc = loc;
       auto last = curr;
       if (stack.empty()) {
         throw ParseException("s-expr stack empty");
@@ -580,14 +582,15 @@ void SExpressionWasmBuilder::parseFunction(Element& s, bool preParseImport) {
   if (importModule.is()) {
     // this is an import, actually
     if (!preParseImport) throw ParseException("!preParseImport in func");
-    std::unique_ptr<Import> im = make_unique<Import>();
+    auto im = make_unique<Function>();
     im->name = name;
     im->module = importModule;
     im->base = importBase;
-    im->kind = ExternalKind::Function;
-    im->functionType = wasm.getFunctionType(type)->name;
-    if (wasm.getImportOrNull(im->name)) throw ParseException("duplicate import", s.line, s.col);
-    wasm.addImport(im.release());
+    im->type = type;
+    FunctionTypeUtils::fillFunction(im.get(), wasm.getFunctionType(type));
+    functionTypes[name] = im->result;
+    if (wasm.getFunctionOrNull(im->name)) throw ParseException("duplicate import", s.line, s.col);
+    wasm.addFunction(im.release());
     if (currFunction) throw ParseException("import module inside function dec");
     currLocalTypes.clear();
     nameMapper.clear();
@@ -608,6 +611,12 @@ void SExpressionWasmBuilder::parseFunction(Element& s, bool preParseImport) {
   if (currFunction->result != result) throw ParseException("bad func declaration", s.line, s.col);
   currFunction->body = body;
   currFunction->type = type;
+  if (s.startLoc) {
+    currFunction->prologLocation.insert(getDebugLocation(*s.startLoc));
+  }
+  if (s.endLoc) {
+    currFunction->epilogLocation.insert(getDebugLocation(*s.endLoc));
+  }
   if (wasm.getFunctionOrNull(currFunction->name)) throw ParseException("duplicate function", s.line, s.col);
   wasm.addFunction(currFunction.release());
   currLocalTypes.clear();
@@ -627,19 +636,23 @@ Type SExpressionWasmBuilder::stringToType(const char* str, bool allowError, bool
   throw ParseException("invalid wasm type");
 }
 
+Function::DebugLocation SExpressionWasmBuilder::getDebugLocation(const SourceLocation& loc) {
+  IString file = loc.filename;
+  auto& debugInfoFileNames = wasm.debugInfoFileNames;
+  auto iter = debugInfoFileIndices.find(file);
+  if (iter == debugInfoFileIndices.end()) {
+    Index index = debugInfoFileNames.size();
+    debugInfoFileNames.push_back(file.c_str());
+    debugInfoFileIndices[file] = index;
+  }
+  uint32_t fileIndex = debugInfoFileIndices[file];
+  return {fileIndex, loc.line, loc.column};
+}
+
 Expression* SExpressionWasmBuilder::parseExpression(Element& s) {
   Expression* result = makeExpression(s);
-  if (s.loc) {
-    IString file = s.loc->filename;
-    auto& debugInfoFileNames = wasm.debugInfoFileNames;
-    auto iter = debugInfoFileIndices.find(file);
-    if (iter == debugInfoFileIndices.end()) {
-      Index index = debugInfoFileNames.size();
-      debugInfoFileNames.push_back(file.c_str());
-      debugInfoFileIndices[file] = index;
-    }
-    uint32_t fileIndex = debugInfoFileIndices[file];
-    currFunction->debugLocations[result] = {fileIndex, s.loc->line, s.loc->column};
+  if (s.startLoc) {
+    currFunction->debugLocations[result] = getDebugLocation(*s.startLoc);
   }
   return result;
 }
@@ -810,7 +823,6 @@ Expression* SExpressionWasmBuilder::makeExpression(Element& s) {
       case 'c': {
         if (str[1] == 'a') {
           if (id == CALL) return makeCall(s);
-          if (id == CALL_IMPORT) return makeCallImport(s);
           if (id == CALL_INDIRECT) return makeCallIndirect(s);
         } else if (str[1] == 'u') return makeHost(s, HostOp::CurrentMemory);
         abort_on(str);
@@ -832,7 +844,6 @@ Expression* SExpressionWasmBuilder::makeExpression(Element& s) {
         abort_on(str);
       }
       case 'h': {
-        if (str[1] == 'a') return makeHost(s, HostOp::HasFeature);
         abort_on(str);
       }
       case 'i': {
@@ -848,7 +859,6 @@ Expression* SExpressionWasmBuilder::makeExpression(Element& s) {
         abort_on(str);
       }
       case 'p': {
-        if (str[1] == 'a') return makeHost(s, HostOp::PageSize);
         abort_on(str);
       }
       case 's': {
@@ -975,11 +985,7 @@ Expression* SExpressionWasmBuilder::makeDrop(Element& s) {
 Expression* SExpressionWasmBuilder::makeHost(Element& s, HostOp op) {
   auto ret = allocator.alloc<Host>();
   ret->op = op;
-  if (op == HostOp::HasFeature) {
-    ret->nameOperand = s[1]->str();
-  } else {
-    parseCallOperands(s, 1, s.size(), ret);
-  }
+  parseCallOperands(s, 1, s.size(), ret);
   if (ret->op == HostOp::GrowMemory) {
     if (ret->operands.size() != 1) {
       throw ParseException("grow_memory needs one operand");
@@ -1035,16 +1041,11 @@ Expression* SExpressionWasmBuilder::makeGetGlobal(Element& s) {
   auto ret = allocator.alloc<GetGlobal>();
   ret->name = getGlobalName(*s[1]);
   auto* global = wasm.getGlobalOrNull(ret->name);
-  if (global) {
-    ret->type = global->type;
-    return ret;
+  if (!global) {
+    throw ParseException("bad get_global name", s.line, s.col);
   }
-  auto* import = wasm.getImportOrNull(ret->name);
-  if (import && import->kind == ExternalKind::Global) {
-    ret->type = import->globalType;
-    return ret;
-  }
-  throw ParseException("bad get_global name", s.line, s.col);
+  ret->type = global->type;
+  return ret;
 }
 
 Expression* SExpressionWasmBuilder::makeSetGlobal(Element& s) {
@@ -1085,6 +1086,9 @@ Expression* SExpressionWasmBuilder::makeBlock(Element& s) {
     if (first[0]->str() == BLOCK) {
       // recurse
       curr = allocator.alloc<Block>();
+      if (first.startLoc) {
+        currFunction->debugLocations[curr] = getDebugLocation(*first.startLoc);
+      }
       sp = &first;
       continue;
     }
@@ -1138,20 +1142,19 @@ Expression* SExpressionWasmBuilder::makeConst(Element& s, Type type) {
   return ret;
 }
 
-static uint8_t parseMemBytes(const char** in, uint8_t fallback) {
+static uint8_t parseMemBytes(const char*& s, uint8_t fallback) {
   uint8_t ret;
-  const char* s = *in;
   if (s[0] == '8') {
     ret = 1;
-    (*in)++;
+    s++;
   } else if (s[0] == '1') {
     if (s[1] != '6') throw ParseException("expected 16 for memop size");
     ret = 2;
-    *in += 2;
+    s += 2;
   } else if (s[0] == '3') {
     if (s[1] != '2') throw ParseException("expected 32 for memop size");;
     ret = 4;
-    *in += 2;
+    s += 2;
   } else {
     ret = fallback;
   }
@@ -1185,13 +1188,23 @@ static size_t parseMemAttributes(Element& s, Address* offset, Address* align, Ad
   return i;
 }
 
+static const char* findMemExtra(const Element& s, size_t skip, bool isAtomic) {
+  auto* str = s.c_str();
+  auto size = strlen(str);
+  auto* ret = strchr(str, '.');
+  if (!ret) throw ParseException("missing '.' in memory access", s.line, s.col);
+  ret += skip;
+  if (isAtomic) ret += 7; // after "type.atomic.load"
+  if (ret > str + size) throw ParseException("memory access ends abruptly", s.line, s.col);
+  return ret;
+}
+
 Expression* SExpressionWasmBuilder::makeLoad(Element& s, Type type, bool isAtomic) {
-  const char *extra = strchr(s[0]->c_str(), '.') + 5; // after "type.load"
-  if (isAtomic) extra += 7; // after "type.atomic.load"
+  const char* extra = findMemExtra(*s[0], 5 /* after "type.load" */, isAtomic);
   auto* ret = allocator.alloc<Load>();
   ret->isAtomic = isAtomic;
   ret->type = type;
-  ret->bytes = parseMemBytes(&extra, getTypeSize(type));
+  ret->bytes = parseMemBytes(extra, getTypeSize(type));
   ret->signed_ = extra[0] && extra[1] == 's';
   size_t i = parseMemAttributes(s, &ret->offset, &ret->align, ret->bytes);
   ret->ptr = parseExpression(s[i]);
@@ -1200,12 +1213,11 @@ Expression* SExpressionWasmBuilder::makeLoad(Element& s, Type type, bool isAtomi
 }
 
 Expression* SExpressionWasmBuilder::makeStore(Element& s, Type type, bool isAtomic) {
-  const char *extra = strchr(s[0]->c_str(), '.') + 6; // after "type.store"
-  if (isAtomic) extra += 7; // after "type.atomic.store"
+  const char* extra = findMemExtra(*s[0], 6 /* after "type.store" */, isAtomic);
   auto ret = allocator.alloc<Store>();
   ret->isAtomic = isAtomic;
   ret->valueType = type;
-  ret->bytes = parseMemBytes(&extra, getTypeSize(type));
+  ret->bytes = parseMemBytes(extra, getTypeSize(type));
   size_t i = parseMemAttributes(s, &ret->offset, &ret->align, ret->bytes);
   ret->ptr = parseExpression(s[i]);
   ret->value = parseExpression(s[i+1]);
@@ -1214,8 +1226,8 @@ Expression* SExpressionWasmBuilder::makeStore(Element& s, Type type, bool isAtom
 }
 
 Expression* SExpressionWasmBuilder::makeAtomicRMWOrCmpxchg(Element& s, Type type) {
-  const char* extra = strchr(s[0]->c_str(), '.') + 11; // afer "type.atomic.rmw"
-  auto bytes = parseMemBytes(&extra, getTypeSize(type));
+  const char* extra = findMemExtra(*s[0], 11 /* after "type.atomic.rmw" */, /* isAtomic = */ false);
+  auto bytes = parseMemBytes(extra, getTypeSize(type));
   extra = strchr(extra, '.'); // after the optional '_u' and before the opcode
   if (!extra) throw ParseException("malformed atomic rmw instruction");
   extra++; // after the '.'
@@ -1361,15 +1373,6 @@ Expression* SExpressionWasmBuilder::makeLoop(Element& s) {
 
 Expression* SExpressionWasmBuilder::makeCall(Element& s) {
   auto target = getFunctionName(*s[1]);
-  auto* import = wasm.getImportOrNull(target);
-  if (import && import->kind == ExternalKind::Function) {
-    auto ret = allocator.alloc<CallImport>();
-    ret->target = target;
-    Import* import = wasm.getImport(ret->target);
-    ret->type = wasm.getFunctionType(import->functionType)->result;
-    parseCallOperands(s, 2, s.size(), ret);
-    return ret;
-  }
   auto ret = allocator.alloc<Call>();
   ret->target = target;
   ret->type = functionTypes[ret->target];
@@ -1378,16 +1381,6 @@ Expression* SExpressionWasmBuilder::makeCall(Element& s) {
   return ret;
 }
 
-Expression* SExpressionWasmBuilder::makeCallImport(Element& s) {
-  auto ret = allocator.alloc<CallImport>();
-  ret->target = s[1]->str();
-  Import* import = wasm.getImport(ret->target);
-  ret->type = wasm.getFunctionType(import->functionType)->result;
-  parseCallOperands(s, 2, s.size(), ret);
-  ret->finalize();
-  return ret;
-}
-
 Expression* SExpressionWasmBuilder::makeCallIndirect(Element& s) {
   if (!wasm.table.exists) throw ParseException("no table");
   auto ret = allocator.alloc<CallIndirect>();
@@ -1517,7 +1510,10 @@ void SExpressionWasmBuilder::stringToBinary(const char* input, size_t size, std:
 
 Index SExpressionWasmBuilder::parseMemoryLimits(Element& s, Index i) {
   wasm.memory.initial = getCheckedAddress(s[i++], "excessive memory init");
-  if (i == s.size()) return i;
+  if (i == s.size()) {
+    wasm.memory.max = Memory::kUnlimitedSize;
+    return i;
+  }
   uint64_t max = atoll(s[i++]->c_str());
   if (max > Memory::kMaxSize) throw ParseException("total memory must be <= 4GB");
   wasm.memory.max = max;
@@ -1527,7 +1523,6 @@ Index SExpressionWasmBuilder::parseMemoryLimits(Element& s, Index i) {
 void SExpressionWasmBuilder::parseMemory(Element& s, bool preParseImport) {
   if (wasm.memory.exists) throw ParseException("too many memories");
   wasm.memory.exists = true;
-  wasm.memory.imported = preParseImport;
   wasm.memory.shared = false;
   Index i = 1;
   if (s[i]->dollared()) {
@@ -1545,15 +1540,8 @@ void SExpressionWasmBuilder::parseMemory(Element& s, bool preParseImport) {
       wasm.addExport(ex.release());
       i++;
     } else if (inner[0]->str() == IMPORT) {
-      importModule = inner[1]->str();
-      importBase = inner[2]->str();
-      auto im = make_unique<Import>();
-      im->kind = ExternalKind::Memory;
-      im->module = importModule;
-      im->base = importBase;
-      im->name = importModule;
-      if (wasm.getImportOrNull(im->name)) throw ParseException("duplicate import", s.line, s.col);
-      wasm.addImport(im.release());
+      wasm.memory.module = inner[1]->str();
+      wasm.memory.base = inner[2]->str();
       i++;
     } else if (inner[0]->str() == "shared") {
       wasm.memory.shared = true;
@@ -1659,70 +1647,69 @@ void SExpressionWasmBuilder::parseExport(Element& s) {
 }
 
 void SExpressionWasmBuilder::parseImport(Element& s) {
-  std::unique_ptr<Import> im = make_unique<Import>();
   size_t i = 1;
   bool newStyle = s.size() == 4 && s[3]->isList(); // (import "env" "STACKTOP" (global $stackTop i32))
+  auto kind = ExternalKind::Invalid;
   if (newStyle) {
     if ((*s[3])[0]->str() == FUNC) {
-      im->kind = ExternalKind::Function;
+      kind = ExternalKind::Function;
     } else if ((*s[3])[0]->str() == MEMORY) {
-      im->kind = ExternalKind::Memory;
+      kind = ExternalKind::Memory;
       if (wasm.memory.exists) throw ParseException("more than one memory");
       wasm.memory.exists = true;
-      wasm.memory.imported = true;
     } else if ((*s[3])[0]->str() == TABLE) {
-      im->kind = ExternalKind::Table;
+      kind = ExternalKind::Table;
       if (wasm.table.exists) throw ParseException("more than one table");
       wasm.table.exists = true;
-      wasm.table.imported = true;
     } else if ((*s[3])[0]->str() == GLOBAL) {
-      im->kind = ExternalKind::Global;
+      kind = ExternalKind::Global;
     } else {
       newStyle = false; // either (param..) or (result..)
     }
   }
   Index newStyleInner = 1;
+  Name name;
   if (s.size() > 3 && s[3]->isStr()) {
-    im->name = s[i++]->str();
+    name = s[i++]->str();
   } else if (newStyle && newStyleInner < s[3]->size() && (*s[3])[newStyleInner]->dollared()) {
-    im->name = (*s[3])[newStyleInner++]->str();
+    name = (*s[3])[newStyleInner++]->str();
   }
-  if (!im->name.is()) {
-    if (im->kind == ExternalKind::Function) {
-      im->name = Name("import$function$" + std::to_string(functionCounter++));
-      functionNames.push_back(im->name);
-    } else if (im->kind == ExternalKind::Global) {
-      im->name = Name("import$global" + std::to_string(globalCounter++));
-      globalNames.push_back(im->name);
-    } else if (im->kind == ExternalKind::Memory) {
-      im->name = Name("import$memory$" + std::to_string(0));
-    } else if (im->kind == ExternalKind::Table) {
-      im->name = Name("import$table$" + std::to_string(0));
+  if (!name.is()) {
+    if (kind == ExternalKind::Function) {
+      name = Name("import$function$" + std::to_string(functionCounter++));
+      functionNames.push_back(name);
+    } else if (kind == ExternalKind::Global) {
+      name = Name("import$global" + std::to_string(globalCounter++));
+      globalNames.push_back(name);
+    } else if (kind == ExternalKind::Memory) {
+      name = Name("import$memory$" + std::to_string(0));
+    } else if (kind == ExternalKind::Table) {
+      name = Name("import$table$" + std::to_string(0));
     } else {
       throw ParseException("invalid import");
     }
   }
   if (!s[i]->quoted()) {
     if (s[i]->str() == MEMORY) {
-      im->kind = ExternalKind::Memory;
+      kind = ExternalKind::Memory;
     } else if (s[i]->str() == TABLE) {
-      im->kind = ExternalKind::Table;
+      kind = ExternalKind::Table;
     } else if (s[i]->str() == GLOBAL) {
-      im->kind = ExternalKind::Global;
+      kind = ExternalKind::Global;
     } else {
       throw ParseException("invalid ext import");
     }
     i++;
   } else if (!newStyle) {
-    im->kind = ExternalKind::Function;
+    kind = ExternalKind::Function;
   }
-  im->module = s[i++]->str();
+  auto module = s[i++]->str();
   if (!s[i]->isStr()) throw ParseException("no name for import");
-  im->base = s[i++]->str();
+  auto base = s[i++]->str();
   // parse internals
   Element& inner = newStyle ? *s[3] : s;
   Index j = newStyle ? newStyleInner : i;
-  if (im->kind == ExternalKind::Function) {
+  if (kind == ExternalKind::Function) {
     std::unique_ptr<FunctionType> type = make_unique<FunctionType>();
     if (inner.size() > j) {
       Element& params = *inner[j];
@@ -1746,27 +1733,46 @@ void SExpressionWasmBuilder::parseImport(Element& s) {
         type->result = stringToType(result[1]->str());
       }
     }
-    im->functionType = ensureFunctionType(getSig(type.get()), &wasm)->name;
-  } else if (im->kind == ExternalKind::Global) {
+    auto func = make_unique<Function>();
+    func->name = name;
+    func->module = module;
+    func->base = base;
+    auto* functionType = ensureFunctionType(getSig(type.get()), &wasm);
+    func->type = functionType->name;
+    FunctionTypeUtils::fillFunction(func.get(), functionType);
+    functionTypes[name] = func->result;
+    wasm.addFunction(func.release());
+  } else if (kind == ExternalKind::Global) {
+    Type type;
     if (inner[j]->isStr()) {
-      im->globalType = stringToType(inner[j]->str());
+      type = stringToType(inner[j]->str());
     } else {
       auto& inner2 = *inner[j];
       if (inner2[0]->str() != MUT) throw ParseException("expected mut");
-      im->globalType = stringToType(inner2[1]->str());
+      type = stringToType(inner2[1]->str());
       throw ParseException("cannot import a mutable global", s.line, s.col);
     }
-  } else if (im->kind == ExternalKind::Table) {
+    auto global = make_unique<Global>();
+    global->name = name;
+    global->module = module;
+    global->base = base;
+    global->type = type;
+    wasm.addGlobal(global.release());
+  } else if (kind == ExternalKind::Table) {
+    wasm.table.module = module;
+    wasm.table.base = base;
     if (j < inner.size() - 1) {
       wasm.table.initial = getCheckedAddress(inner[j++], "excessive table init size");
     }
     if (j < inner.size() - 1) {
       wasm.table.max = getCheckedAddress(inner[j++], "excessive table max size");
     } else {
-      wasm.table.max = Table::kMaxSize;
+      wasm.table.max = Table::kUnlimitedSize;
     }
     // ends with the table element type
-  } else if (im->kind == ExternalKind::Memory) {
+  } else if (kind == ExternalKind::Memory) {
+    wasm.memory.module = module;
+    wasm.memory.base = base;
     if (inner[j]->isList()) {
       auto& limits = *inner[j];
       if (!(limits[0]->isStr() && limits[0]->str() == "shared")) throw ParseException("bad memory limit declaration");
@@ -1776,8 +1782,6 @@ void SExpressionWasmBuilder::parseImport(Element& s) {
       parseMemoryLimits(inner, j);
     }
   }
-  if (wasm.getImportOrNull(im->name)) throw ParseException("duplicate import", s.line, s.col);
-  wasm.addImport(im.release());
 }
 
 void SExpressionWasmBuilder::parseGlobal(Element& s, bool preParseImport) {
@@ -1825,14 +1829,13 @@ void SExpressionWasmBuilder::parseGlobal(Element& s, bool preParseImport) {
     // this is an import, actually
     if (!preParseImport) throw ParseException("!preParseImport in global");
     if (mutable_) throw ParseException("cannot import a mutable global", s.line, s.col);
-    std::unique_ptr<Import> im = make_unique<Import>();
+    auto im = make_unique<Global>();
     im->name = global->name;
     im->module = importModule;
     im->base = importBase;
-    im->kind = ExternalKind::Global;
-    im->globalType = type;
-    if (wasm.getImportOrNull(im->name)) throw ParseException("duplicate import", s.line, s.col);
-    wasm.addImport(im.release());
+    im->type = type;
+    if (wasm.getGlobalOrNull(im->name)) throw ParseException("duplicate import", s.line, s.col);
+    wasm.addGlobal(im.release());
     return;
   }
   if (preParseImport) throw ParseException("preParseImport in global");
@@ -1852,7 +1855,6 @@ void SExpressionWasmBuilder::parseGlobal(Element& s, bool preParseImport) {
 void SExpressionWasmBuilder::parseTable(Element& s, bool preParseImport) {
   if (wasm.table.exists) throw ParseException("more than one table");
   wasm.table.exists = true;
-  wasm.table.imported = preParseImport;
   Index i = 1;
   if (i == s.size()) return; // empty table in old notation
   if (s[i]->dollared()) {
@@ -1871,16 +1873,9 @@ void SExpressionWasmBuilder::parseTable(Element& s, bool preParseImport) {
       wasm.addExport(ex.release());
       i++;
     } else if (inner[0]->str() == IMPORT) {
-      importModule = inner[1]->str();
-      importBase = inner[2]->str();
       if (!preParseImport) throw ParseException("!preParseImport in table");
-      auto im = make_unique<Import>();
-      im->kind = ExternalKind::Table;
-      im->module = importModule;
-      im->base = importBase;
-      im->name = importModule;
-      if (wasm.getImportOrNull(im->name)) throw ParseException("duplicate import", s.line, s.col);
-      wasm.addImport(im.release());
+      wasm.table.module = inner[1]->str();
+      wasm.table.base = inner[2]->str();
       i++;
     } else {
       throw ParseException("invalid table");
