@@ -29,6 +29,8 @@ namespace BinaryConsts {
 namespace UserSections {
 const char* Name = "name";
 const char* SourceMapUrl = "sourceMappingURL";
+
+const char* Dylink = "dylink";
 }
 }
 
@@ -51,7 +53,6 @@ Name GROW_WASM_MEMORY("__growWasmMemory"),
      LOCAL("local"),
      TYPE("type"),
      CALL("call"),
-     CALL_IMPORT("call_import"),
      CALL_INDIRECT("call_indirect"),
      BLOCK("block"),
      BR_IF("br_if"),
@@ -81,7 +82,6 @@ const char* getExpressionName(Expression* curr) {
     case Expression::Id::BreakId: return "break";
     case Expression::Id::SwitchId: return "switch";
     case Expression::Id::CallId: return "call";
-    case Expression::Id::CallImportId: return "call_import";
     case Expression::Id::CallIndirectId: return "call_indirect";
     case Expression::Id::GetLocalId: return "get_local";
     case Expression::Id::SetLocalId: return "set_local";
@@ -322,10 +322,6 @@ void Call::finalize() {
   handleUnreachableOperands(this);
 }
 
-void CallImport::finalize() {
-  handleUnreachableOperands(this);
-}
-
 void CallIndirect::finalize() {
   handleUnreachableOperands(this);
   if (target->type == unreachable) {
@@ -556,7 +552,7 @@ void Drop::finalize() {
 
 void Host::finalize() {
   switch (op) {
-    case PageSize: case CurrentMemory: case HasFeature: {
+    case CurrentMemory: {
       type = i32;
       break;
     }
@@ -619,8 +615,11 @@ Name Function::getLocalNameOrGeneric(Index index) {
 }
 
 Index Function::getLocalIndex(Name name) {
-  assert(localIndices.count(name) > 0);
-  return localIndices[name];
+  auto iter = localIndices.find(name);
+  if (iter == localIndices.end()) {
+    Fatal() << "Function::getLocalIndex: " << name << " does not exist";
+  }
+  return iter->second;
 }
 
 Index Function::getVarIndexBase() {
@@ -638,92 +637,110 @@ Type Function::getLocalType(Index index) {
 }
 
 FunctionType* Module::getFunctionType(Name name) {
-  assert(functionTypesMap.count(name));
-  return functionTypesMap[name];
-}
-
-Import* Module::getImport(Name name) {
-  assert(importsMap.count(name));
-  return importsMap[name];
+  auto iter = functionTypesMap.find(name);
+  if (iter == functionTypesMap.end()) {
+    Fatal() << "Module::getFunctionType: " << name << " does not exist";
+  }
+  return iter->second;
 }
 
 Export* Module::getExport(Name name) {
-  assert(exportsMap.count(name));
-  return exportsMap[name];
+  auto iter = exportsMap.find(name);
+  if (iter == exportsMap.end()) {
+    Fatal() << "Module::getExport: " << name << " does not exist";
+  }
+  return iter->second;
 }
 
 Function* Module::getFunction(Name name) {
-  assert(functionsMap.count(name));
-  return functionsMap[name];
+  auto iter = functionsMap.find(name);
+  if (iter == functionsMap.end()) {
+    Fatal() << "Module::getFunction: " << name << " does not exist";
+  }
+  return iter->second;
 }
 
 Global* Module::getGlobal(Name name) {
-  assert(globalsMap.count(name));
-  return globalsMap[name];
+  auto iter = globalsMap.find(name);
+  if (iter == globalsMap.end()) {
+    Fatal() << "Module::getGlobal: " << name << " does not exist";
+  }
+  return iter->second;
 }
 
 FunctionType* Module::getFunctionTypeOrNull(Name name) {
-  if (!functionTypesMap.count(name))
+  auto iter = functionTypesMap.find(name);
+  if (iter == functionTypesMap.end()) {
     return nullptr;
-  return functionTypesMap[name];
-}
-
-Import* Module::getImportOrNull(Name name) {
-  if (!importsMap.count(name))
-    return nullptr;
-  return importsMap[name];
+  }
+  return iter->second;
 }
 
 Export* Module::getExportOrNull(Name name) {
-  if (!exportsMap.count(name))
+  auto iter = exportsMap.find(name);
+  if (iter == exportsMap.end()) {
     return nullptr;
-  return exportsMap[name];
+  }
+  return iter->second;
 }
 
 Function* Module::getFunctionOrNull(Name name) {
-  if (!functionsMap.count(name))
+  auto iter = functionsMap.find(name);
+  if (iter == functionsMap.end()) {
     return nullptr;
-  return functionsMap[name];
+  }
+  return iter->second;
 }
 
 Global* Module::getGlobalOrNull(Name name) {
-  if (!globalsMap.count(name))
+  auto iter = globalsMap.find(name);
+  if (iter == globalsMap.end()) {
     return nullptr;
-  return globalsMap[name];
+  }
+  return iter->second;
 }
 
 void Module::addFunctionType(FunctionType* curr) {
-  assert(curr->name.is());
+  if (!curr->name.is()) {
+    Fatal() << "Module::addFunctionType: empty name";
+  }
+  if (getFunctionTypeOrNull(curr->name)) {
+    Fatal() << "Module::addFunctionType: " << curr->name << " already exists";
+  }
   functionTypes.push_back(std::unique_ptr<FunctionType>(curr));
-  assert(functionTypesMap.find(curr->name) == functionTypesMap.end());
   functionTypesMap[curr->name] = curr;
 }
 
-void Module::addImport(Import* curr) {
-  assert(curr->name.is());
-  imports.push_back(std::unique_ptr<Import>(curr));
-  assert(importsMap.find(curr->name) == importsMap.end());
-  importsMap[curr->name] = curr;
-}
-
 void Module::addExport(Export* curr) {
-  assert(curr->name.is());
+  if (!curr->name.is()) {
+    Fatal() << "Module::addExport: empty name";
+  }
+  if (getExportOrNull(curr->name)) {
+    Fatal() << "Module::addExport: " << curr->name << " already exists";
+  }
   exports.push_back(std::unique_ptr<Export>(curr));
-  assert(exportsMap.find(curr->name) == exportsMap.end());
   exportsMap[curr->name] = curr;
 }
 
 void Module::addFunction(Function* curr) {
-  assert(curr->name.is());
+  if (!curr->name.is()) {
+    Fatal() << "Module::addFunction: empty name";
+  }
+  if (getFunctionOrNull(curr->name)) {
+    Fatal() << "Module::addFunction: " << curr->name << " already exists";
+  }
   functions.push_back(std::unique_ptr<Function>(curr));
-  assert(functionsMap.find(curr->name) == functionsMap.end());
   functionsMap[curr->name] = curr;
 }
 
 void Module::addGlobal(Global* curr) {
-  assert(curr->name.is());
+  if (!curr->name.is()) {
+    Fatal() << "Module::addGlobal: empty name";
+  }
+  if (getGlobalOrNull(curr->name)) {
+    Fatal() << "Module::addGlobal: " << curr->name << " already exists";
+  }
   globals.push_back(std::unique_ptr<Global>(curr));
-  assert(globalsMap.find(curr->name) == globalsMap.end());
   globalsMap[curr->name] = curr;
 }
 
@@ -731,14 +748,14 @@ void Module::addStart(const Name& s) {
   start = s;
 }
 
-void Module::removeImport(Name name) {
-  for (size_t i = 0; i < imports.size(); i++) {
-    if (imports[i]->name == name) {
-      imports.erase(imports.begin() + i);
+void Module::removeFunctionType(Name name) {
+  for (size_t i = 0; i < functionTypes.size(); i++) {
+    if (functionTypes[i]->name == name) {
+      functionTypes.erase(functionTypes.begin() + i);
       break;
     }
   }
-  importsMap.erase(name);
+  functionTypesMap.erase(name);
 }
 
 void Module::removeExport(Name name) {
@@ -761,14 +778,14 @@ void Module::removeFunction(Name name) {
   functionsMap.erase(name);
 }
 
-void Module::removeFunctionType(Name name) {
-  for (size_t i = 0; i < functionTypes.size(); i++) {
-    if (functionTypes[i]->name == name) {
-      functionTypes.erase(functionTypes.begin() + i);
+void Module::removeGlobal(Name name) {
+  for (size_t i = 0; i < globals.size(); i++) {
+    if (globals[i]->name == name) {
+      globals.erase(globals.begin() + i);
       break;
     }
   }
-  functionTypesMap.erase(name);
+  globalsMap.erase(name);
 }
 
 // TODO: remove* for other elements
@@ -782,10 +799,6 @@ void Module::updateMaps() {
   for (auto& curr : functionTypes) {
     functionTypesMap[curr->name] = curr.get();
   }
-  importsMap.clear();
-  for (auto& curr : imports) {
-    importsMap[curr->name] = curr.get();
-  }
   exportsMap.clear();
   for (auto& curr : exports) {
     exportsMap[curr->name] = curr.get();
